"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/datamodel.py
"""

import abc


class AbstractDataModel(abc.ABC):
    """
    This Abstract Base Class is intended to cover multiple implementations of
    data models so that each will be considered an appropriate subclass of this
    class without requiring that they inherit this class.

    Any datamodel class instance that desires to be considered an instance of
    AbstractDataModel must implement the following methods.

    In addition, although it isn't yet checked (the best approach for supporting
    this is still being considered), such instances must have a meta.filename
    attribute.
    """

    @classmethod
    def __subclasshook__(cls, c_):
        """
        Pseudo subclass check based on these attributes and methods
        """
        if cls is AbstractDataModel:
            mro = c_.__mro__
            if (
                any(hasattr(CC, "crds_observatory") for CC in mro)
                and any(hasattr(CC, "get_crds_parameters") for CC in mro)
                and any(hasattr(CC, "save") for CC in mro)
            ):
                return True
        return False

    @property
    @abc.abstractmethod
    def crds_observatory(self):
        """This should return a string identifying the observatory as CRDS expects it"""

    @abc.abstractmethod
    def get_crds_parameters(self):
        """
        This should return a dictionary of key/value pairs corresponding to the
        parkey values CRDS is using to match reference files. Typically it returns
        all metadata simple values.
        """

    @abc.abstractmethod
    def save(self, path, dir_path=None, *args, **kwargs):
        """
        Save to a file.

        Parameters
        ----------
        path : string or callable
            File path to save to.
            If function, it takes one argument that is
            model.meta.filename and returns the full path string.

        dir_path : str
            Directory to save to. If not None, this will override
            any directory information in the ``path``

        Returns
        -------
        output_path: str
            The file path the model was saved in.
        """


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/model_blender/rules.py
"""

import numpy as np


__all__ = ['RULE_FUNCTIONS', 'AttributeBlender', 'make_blender']


def _multi(vals):
    """
    This will either return the common value from a list of identical values
    or 'MULTIPLE'
    """
    uniq_vals = list(set(vals))
    num_vals = len(uniq_vals)
    if num_vals == 0:
        return None
    if num_vals == 1:
        return uniq_vals[0]
    if num_vals > 1:
        return "MULTIPLE"


RULE_FUNCTIONS = {
    'multi': _multi,
    'mean': np.mean,
    'sum': np.sum,
    'max': np.max,
    'min': np.min,

     # retained date/time names for backwards compatibility
     # as these all assume ISO8601 format the lexical and
     # chronological sorting match
    'mintime': min,
    'maxtime': max,
    'mindate': min,
    'maxdate': max,
    'mindatetime': min,
    'maxdatetime': max
}
"""
Mapping of rule names to functions.

Used for `make_blender`.

The following rules are considered deprecated
and should not be used for new schemas.

  - mintime
  - maxtime
  - mindate
  - maxdate
  - mindatetime
  - maxdatetime
"""


class AttributeBlender:
    """
    Single attribute metadata blender
    """
    def __init__(self, blend_function):
        """
        Create a new metadata attribute blender.

        Parameters
        ----------
        blend_function: callable
            Function to blend accumulated metadata values
        """
        self.blend_function = blend_function
        self.values = []

    def accumulate(self, value):
        """
        Add a metadata value for blending.

        Parameters
        ----------
        value:
            Value for this metadata attribute to use
            when blending.
        """
        self.values.append(value)

    def finalize(self):
        """
        Blend the accumulated metadata values.

        Returns
        -------
        value:
            The blended result.
        """
        if not self.values:
            return None
        return self.blend_function(self.values)


def make_blender(rule):
    """
    Make a `AttributeBlender` instance using the provided rule

    Parameters
    ----------
    rule: string
        Name of the blending rule. Must be in `RULE_FUNCTIONS`.

    Returns
    -------
    attr_blender: `AttrBlender`
        Blender instance using the provided rule.
    """
    return AttributeBlender(RULE_FUNCTIONS[rule])


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/lib/keyvalue_registry.py
"""

from collections import UserDict

__all__ = [
    "KeyValueRegistry",
    "KeyValueRegistryError",
    "KeyValueRegistryNoKeyFound",
    "KeyValueRegistryNotSingleItemError",
]


class KeyValueRegistry(UserDict):
    """Provide a dict-like registry

    Differences from just a `dict`:
        - Can be given single item or a 2-tuple.
          If an item, attempts to read the `__name__` attribute
          and use that as the key.

        - If None is given as a key, a default key can
          be specified.

        - Instances can be used as decorators.

    Parameters
    ----------
    items : object or (str, object) or dict
        Initializing items.

    default : str or object
        The default to use when key is `None`
    """

    def __init__(self, items=None, default=None):
        super_args = ()
        if items is not None:
            super_args = (make_dict(items),)
        super().__init__(*super_args)

        self.default = None
        if default is not None:
            default_dict = make_dict(default)
            if len(default_dict) > 1:
                raise KeyValueRegistryNotSingleItemError
            default_dict = make_dict(default)
            self.update(default_dict)
            self.default = next(iter(default_dict.keys()))
            self.update({None: default_dict[self.default]})

    def update(self, item):
        """Add item to registry"""
        item_dict = make_dict(item)
        super().update(item_dict)

    def __call__(self, item):
        """Add item by calling instance

        This allows an instance to be used as a decorator.
        """
        self.update(item)
        return item


# ******
# Errors
# ******
class KeyValueRegistryError(Exception):
    def __init__(self, *args):
        if len(args) == 0:
            args = (self.msg,)
        super().__init__(*args)


class KeyValueRegistryNotSingleItemError(KeyValueRegistryError):
    msg = "Item cannot be a list"


class KeyValueRegistryNoKeyFound(KeyValueRegistryError):
    msg = "Cannot deduce key from given value"


# *********
# Utilities
# *********
def make_dict(item):
    """Create a dict from an item

    Parameters
    ----------
    item : object or (name, object) or dict
        If dict, just return dict.
        If 2-tuple, return dict with the key/value pair
        If just object, use `__name__` as key
    """
    try:
        item_dict = dict(item)
    except (TypeError, ValueError):
        try:
            key, value = item
        except (TypeError, ValueError):
            try:
                key = item.__name__
            except (AttributeError, SyntaxError):
                raise KeyValueRegistryNoKeyFound
            else:
                value = item

        # At they point we have a key/value pair
        item_dict = {key: value}

    return item_dict


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/utilities.py
"""

import inspect
import os
import sys
import types

from spike.stpipe import entry_points

def resolve_step_class_alias(name):
    """
    If the input is a recognized alias, return the
    corresponding fully-qualified class name.  Otherwise
    return the input unmodified.

    Parameters
    ----------
    name : str
        If name contains "::" only the package with
        a name matching the characters before "::"
        will be searched for the matching step.

    Returns
    -------
    str
    """
    # check if the name contains a package name
    if "::" in name:
        scope, class_name = name.split("::", maxsplit=1)
    else:
        scope, class_name = None, name

    # track all found steps keyed by package name
    found_class_names = {}
    for info in entry_points.get_steps():
        if scope and info.package_name != scope:
            continue
        if info.class_alias is not None and class_name == info.class_alias:
            found_class_names[info.package_name] = info

    if not found_class_names:
        return name

    if len(found_class_names) == 1:
        return found_class_names.popitem()[1].class_name

    # class alias resolved to several possible steps
    scopes = list(found_class_names.keys())
    msg = (
        f"class alias {name} matched more than 1 step. Please provide "
        "the package name along with the step name. One of:\n"
    )
    for scope in scopes:
        msg += f"  {scope}::{name}\n"
    raise ValueError(msg)


def import_class(full_name, subclassof=object, config_file=None):
    """
    Import the Python class `full_name` given in full Python package format,
    e.g.::

        package.another_package.class_name

    Return the imported class. Optionally, if `subclassof` is not None
    and is a Python class, make sure that the imported class is a
    subclass of `subclassof`.
    """
    # Understand which class we need to instantiate. The class name is given in
    # full Python package notation, e.g.
    #   package.subPackage.subsubpackage.className
    # in the input parameter `full_name`. This means that
    #   1. We HAVE to be able to say
    #       `from package.subPackage.subsubpackage import className`
    #   2. If `subclassof` is defined, the newly imported Python class MUST be a
    #      subclass of `subclassof`, which HAS to be a Python class.

    if config_file is not None:
        sys.path.insert(0, os.path.dirname(config_file))

    try:
        full_name = full_name.strip()
        package_name, _, class_name = full_name.rpartition(".")
        if not package_name:
            raise ImportError(f"{full_name} is not a Python class")
        imported = __import__(
            package_name,
            globals(),
            locals(),
            [
                class_name,
            ],
            level=0,
        )

        step_class = getattr(imported, class_name)

        if not isinstance(step_class, type):
            raise TypeError(
                f"Object {class_name} from package {package_name} is not a class"
            )

        if not issubclass(step_class, subclassof):
            raise TypeError(
                f"Class {class_name} from package {package_name} is not a subclass of"
                f" {subclassof.__name__}"
            )
    finally:
        if config_file is not None:
            del sys.path[0]

    return step_class


def import_func(full_name):
    """
    Import the Python class `full_name` given in full Python package format,
    e.g.::

        package.subpackage.subsubpackage.function_name

    Return the imported function.
    """
    full_name = full_name.strip()
    package_name, _, func_name = full_name.rpartition(".")
    if not package_name:
        raise ImportError(f"{full_name} is not a fully qualified path to function")
    imported = __import__(
        package_name,
        globals(),
        locals(),
        [
            func_name,
        ],
        level=0,
    )

    step_func = getattr(imported, func_name)

    if not isinstance(step_func, types.FunctionType):
        raise TypeError(
            f"Object {func_name} from package {package_name} is not a function"
        )

    return step_func


def get_spec_file_path(step_class):
    """
    Given a Step (sub)class, divine and return the full path to the
    corresponding spec file. Use the fact that by convention, the spec
    file is in the same directory as the `step_class` source file. It
    has the name of the Step (sub)class and extension .spec.
    """
    try:
        step_source_file = inspect.getfile(step_class)
    except TypeError:
        return None
    step_source_file = os.path.abspath(step_source_file)

    # Since `step_class` could be defined in a file called whatever,
    # we need the source file basedir and the class name.
    dir_ = os.path.dirname(step_source_file)
    return os.path.join(dir_, step_class.__name__ + ".spec")


def find_spec_file(step_class):
    """
    Return the full path of the given Step subclass `step_class`, if
    it exists or None if it does not.
    """
    spec_file = get_spec_file_path(step_class)
    if spec_file is not None and os.path.exists(spec_file):
        return spec_file
    return None


def get_fully_qualified_class_name(cls_or_obj):
    """
    Return an object's fully-qualified class name.

    Parameters
    ----------
    cls_or_obj : type or object

    Returns
    -------
    str
        For type input, the fully-qualified name of the type.
        For other input, the fully-qualified name of the input's class.
    """
    cls = cls_or_obj if inspect.isclass(cls_or_obj) else cls_or_obj.__class__
    module = cls.__module__
    if module is None or module == str.__class__.__module__:
        return cls.__name__  # Avoid reporting __builtin__

    return module + "." + cls.__name__


class _NotSet:
    """
    Special value indicating that a parameter is not set.  Distinct
    from None. Instead of using this class use the _not_set instance
    below
    """


_not_set = _NotSet()


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/stpipe/core.py
"""

import importlib.metadata
import logging
import time
from pathlib import Path

import roman_datamodels as rdm
from roman_datamodels.datamodels import ImageModel, MosaicModel
# from stpipe import Pipeline, Step, crds_client
from spike.stpipe.step import Step
from spike.stpipe.pipeline import Pipeline
from spike.stpipe import crds_client

from spike.romancal.datamodels_library import ModelLibrary

from spike.romancal.lib_suffix import remove_suffix

_LOG_FORMATTER = logging.Formatter(
    "%(asctime)s.%(msecs)03dZ :: %(name)s :: %(levelname)s :: %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)
_LOG_FORMATTER.converter = time.gmtime


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanStep(Step):
    """
    Base class for Roman calibration pipeline steps.
    """

    spec = """
    output_ext =  string(default='.asdf')    # Default type of output
    """

    _log_records_formatter = _LOG_FORMATTER

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        """
        Provide access to this package's datamodels.open function
        so that the stpipe infrastructure knows how to instantiate
        models and containers.
        """
        if isinstance(init, str):
            init = Path(init)
        if isinstance(init, Path):
            ext = init.suffix.lower()
            if ext == ".asdf":
                return rdm.open(init, **kwargs)
            if ext in (".json", ".yaml"):
                return ModelLibrary(init, **kwargs)
        if isinstance(init, rdm.DataModel):
            return rdm.open(init, **kwargs)
        if isinstance(init, ModelLibrary):
            return ModelLibrary(init)
        raise TypeError(f"Invalid input: {init}")

    def finalize_result(self, model, reference_files_used):
        """
        Hook that allows the Step to set metadata on the output model
        before save.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.

        reference_files_used : list of tuple(str, str)
            List of reference files used.  The first element of each tuple
            is the reftype code, the second element is the filename.
        """

        model.meta.calibration_software_version = importlib.metadata.version("romancal")

        # FIXME: should cal_logs be brought back into meta for a MosaicModel?
        if isinstance(model, ImageModel):
            # convert to model.cal_logs type to avoid validation errors
            model.meta.cal_logs = type(model.meta.cal_logs)(self.log_records)
        if isinstance(model, MosaicModel):
            model.cal_logs = type(model.cal_logs)(self.log_records)

        if len(reference_files_used) > 0:
            for ref_name, ref_file in reference_files_used:
                if hasattr(model.meta.ref_file, ref_name):
                    setattr(model.meta.ref_file, ref_name, ref_file)
                    # getattr(model.meta.ref_file, ref_name).name = ref_file
            model.meta.ref_file.crds.version = crds_client.get_svn_version()
            model.meta.ref_file.crds.context = crds_client.get_context_used(
                model.crds_observatory
            )

            # this will only run if 'parent' is none, which happens when an individual
            # step is being run or if self is a RomanPipeline and not a RomanStep.
            if self.parent is None:
                log.info(
                    f"Results used CRDS context: {model.meta.ref_file.crds.context}"
                )

    def record_step_status(self, model, step_name, success=True):
        """
        Record step completion status in the model's metadata.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.
        step_name : str
            Calibration step name.
        success : bool
            If True, then the step was run successfully.
        """
        # JWST sets model.meta.cal_step.<step name> here.  Roman
        # may do the same, depending on how the metadata format
        # turns out.  Seems like we might be able to combine this
        # with finalize_result somehow.
        pass

    def remove_suffix(self, name):
        """
        Remove any Roman step-specific suffix from the given filename.

        Parameters
        ----------
        name : str
            Filename.

        Returns
        -------
        str
            Filename with step suffix removed.
        """
        # JWST maintains a list of relevant suffixes that is monitored
        # by tests to be up-to-date.  Roman will likely need to do
        # something similar.
        return remove_suffix(name)


# RomanPipeline needs to inherit from Pipeline, but also
# be a subclass of RomanStep so that it will pass checks
# when constructing a pipeline using RomanStep class methods.
class RomanPipeline(Pipeline, RomanStep):
    pass


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/asn_from_list.py
"""

import argparse
import sys

from spike.jwstcal.associations_registry import AssociationRegistry
from spike.jwstcal.associations_lib_rules_level3_base import DMS_Level3_Base

__all__ = ['asn_from_list']


def asn_from_list(items, rule=DMS_Level3_Base, **kwargs):
    """Create an association from a list

    Parameters
    ----------
    items: [object [, ...]]
        List of items to add.

    rule: `Association` rule
        The association rule to use.

    kwargs: dict
        Other named parameters required or pertinent to adding
        the items to the association.

    Returns
    -------
    association: `Association`-based instance
        The association with the items added.

    Notes
    -----
    This is a lower-level tool for artificially creating
    an association. As such, the association created may not be valid.
    It is presumed the user knows what they are doing.
    """

    asn = rule()
    asn._add_items(items, **kwargs)
    return asn


class Main():
    """Command-line interface for list_to_asn

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """
    def __init__(self, args=None):
        self.configure(args)

    def asn_from_list(self):
        """Create the associaton from the list"""
        parsed = self.parsed
        self.asn = asn_from_list(
            parsed.filelist,
            rule=self.rule,
            product_name=parsed.product_name,
            acid=parsed.acid
        )

    @classmethod
    def cli(cls, args=None):
        """Command-line interface to creating an association from a list

        Parameters
        ----------
        args: [str, ...], or None
            The command line arguments. Can be one of
                - `None`: `sys.argv` is then used.
                - `[str, ...]`: A list of strings which create the command line
                  with the similar structure as `sys.argv`
        """
        association = cls(args=args)
        association.asn_from_list()
        association.save()
        return association

    def configure(self, args=None):
        """Configure

        Parameters
        ----------
        args: [str, ...], or None
            The command line arguments. Can be one of
                - `None`: `sys.argv` is then used.
                - `[str, ...]`: A list of strings which create the command line
                  with the similar structure as `sys.argv`
        """
        if args is None:
            args = sys.argv[1:]
        if isinstance(args, str):
            args = args.split(' ')

        parser = argparse.ArgumentParser(
            description='Create an association from a list of files',
            usage='asn_from_list -o mosaic_asn.json --product-name my_mosaic *.fits'
        )

        parser.add_argument(
            '-o', '--output-file',
            type=str,
            required=True,
            help='File to write association to'
        )

        parser.add_argument(
            '-f', '--format',
            type=str,
            default='json',
            help='Format of the association files. Default: "%(default)s"'
        )

        parser.add_argument(
            '--product-name',
            type=str,
            help='The product name when creating a Level 3 association'
        )

        parser.add_argument(
            '-r', '--rule',
            type=str,
            default='DMS_Level3_Base',
            help=(
                'The rule to base the association structure on.'
                ' Default: "%(default)s"'
            )
        )
        parser.add_argument(
            '--ruledefs', action='append',
            help=(
                'Association rules definition file(s)'
                ' If not specified, the default rules will be searched.'
            )
        )
        parser.add_argument(
            '-i', '--id',
            type=str,
            default='o999',
            help='The association candidate id to use. Default: "%(default)s"',
            dest='acid'
        )

        parser.add_argument(
            'filelist',
            type=str,
            nargs='+',
            help='File list to include in the association'
        )

        self.parsed = parsed = parser.parse_args(args=args)

        # Get the rule
        self.rule = AssociationRegistry(parsed.ruledefs, include_bases=True)[parsed.rule]

    def save(self):
        """Save association"""
        parsed = self.parsed
        with open(parsed.output_file, 'w') as outfile:
            name, serialized = self.asn.dump(format=parsed.format)
            outfile.write(serialized)


def main(args=None):
    """Command-line entrypoint for asn_from_list

    Wrapper around `Main.cli` so that the return is either True or an exception

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """
    Main.cli(args=args)


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/lib/acid.py
"""

import re
from ast import literal_eval

from spike.romancal.associations_lib_counter import Counter

__all__ = ["ACID"]

# Start of the discovered association ids.
_DISCOVERED_ID_START = 3001


class ACID:
    """Association Candidate Identifer

    Parameters
    ----------
    input : str or 2-tuple
        The string representation or 2-tuple containing the
        candidate ID and TYPE. The string should be itself
        the 2-tuple representation when evaluated. The 2-tuple
        will have the form:
            (id, type)

    Attributes
    ----------
    id : str
        The id number.

    type : str
        The type of candidate. Some, but not all
        possibilities include 'OBSERVATION',
        'MOSAIC', 'DISCOVERED'

    __str__ : str
        The DMS-specified string representation of
        a candidate identifier. Possibilities include
        'oXXX' for OBSERVATION, 'c1XXX' for 'MOSAIC' or
        other PPS-defined candidates, and 'a3XXX' for
        'DISCOVERED' associations.
    """

    def __init__(self, input):
        try:
            self.id, self.type = literal_eval(input)
        except (ValueError, SyntaxError):
            self.id, self.type = input

    def __str__(self):
        return self.id


class ACIDMixin:
    """Enable ACID for rules"""

    def __init__(self, *args, **kwargs):
        # Initialize discovered association ID
        self.discovered_id = Counter(_DISCOVERED_ID_START)

        super().__init__(*args, **kwargs)

    def acid_from_constraints(self):
        """Determine ACID from constraints"""
        for constraint in self.constraints:
            if getattr(constraint, "is_acid", False):
                value = re.sub("\\\\", "", "-".join(constraint.found_values))
                try:
                    acid = ACID(value)
                except ValueError:
                    pass
                else:
                    break
        else:
            id = f"a{self.discovered_id.value:0>3}"
            acid = ACID((id, "DISCOVERED"))

        return acid



"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/lib/pipe_utils.py
"""

import logging

import numpy as np
from stdatamodels.properties import ObjectNode
from stdatamodels.jwst.datamodels import dqflags, JwstDataModel

from spike.jwstcal.associations_lib_dms_base import TSO_EXP_TYPES


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def is_tso(model):
    """Is data Time Series Observation data?

    Parameters
    ----------
    model : `~jwst.datamodels.JwstDataModel`
        Data to check

    Returns
    -------
    is_tso : bool
       `True` if the model represents TSO data
    """
    is_tso = False

    # Check on JWST-specific TSOVISIT flag
    try:
        is_tso = model.meta.visit.tsovisit
    except AttributeError:
        pass

    # Check on exposure types
    try:
        is_tso = is_tso or model.meta.exposure.type.lower() in TSO_EXP_TYPES
    except AttributeError:
        pass

    # Check on number of integrations
    try:
        if model.meta.exposure.nints is not None and model.meta.exposure.nints < 2:
            is_tso = False
    except AttributeError:
        pass

    # We've checked everything.
    return is_tso


def is_irs2(model):
    """Check whether the data are in IRS2 format.

    This currently assumes that only full-frame, near-infrared data can be
    taken using the IRS2 readout pattern.

    Parameters
    ----------
    model : `~jwst.datamodels.JwstDataModel` or ndarray
        Data to check

    Returns
    -------
    bool
       `True` if the data are in IRS2 format
    """

    if isinstance(model, np.ndarray):
        shape = model.shape
    else:
        try:
            shape = model.data.shape
        except AttributeError:
            return False

    max_length = 2048

    irs2_axis_length = max(shape[-1], shape[-2])

    if irs2_axis_length > max_length:
        return True
    else:
        return False


def match_nans_and_flags(input_model):
    """Ensure data, error, variance, and DQ are marked consistently for invalid data.

    Invalid data is assumed to be any pixel set to NaN in any one of the
    data, error, or variance arrays, or else set to the DO_NOT_USE flag
    in the DQ array.

    The input model is updated in place with NaNs or DO_NOT_USE flags, as
    appropriate, at all invalid data locations.

    Parameters
    ----------
    input_model : DataModel
        Input model containing some combination of data, dq, err, var_rnoise,
        var_poisson, and var_flat extensions. These extensions must all have
        matching dimensions if present.
    """
    # Check for datamodel input or slit instance
    if (not isinstance(input_model, JwstDataModel)
            and not isinstance(input_model, ObjectNode)):
        raise ValueError(f"Input {type(input_model)} is not a datamodel.")

    # Build up the invalid data flags from each available data extension.
    is_invalid = None
    data_shape = None
    nan_extensions = ['data', 'err', 'var_rnoise', 'var_poisson', 'var_flat']
    for extension in nan_extensions:
        if not hasattr(input_model, extension):
            continue
        data = getattr(input_model, extension)
        if is_invalid is None:
            is_invalid = np.isnan(data)
            data_shape = data.shape
        else:
            if data.shape != data_shape:
                log.warning(f"Mismatched data shapes; skipping invalid data "
                            f"updates for extension '{extension}'")
                continue
            is_invalid |= np.isnan(data)

    # Nothing to do if no extensions were found to update
    if is_invalid is None:
        return

    # Add in invalid flags from the DQ extension if present
    if hasattr(input_model, 'dq'):
        do_not_use = (input_model.dq & dqflags.pixel['DO_NOT_USE']).astype(bool)
        if input_model.dq.shape != data_shape:
            log.warning("Mismatched data shapes; skipping invalid data "
                        "updates for extension 'dq'")
        else:
            is_invalid |= do_not_use

    # Update all the data extensions
    for extension in nan_extensions:
        if not hasattr(input_model, extension):
            continue
        data = getattr(input_model, extension)
        if data.shape != data_shape:
            continue
        data[is_invalid] = np.nan

    # Update the DQ extension
    if input_model.dq.shape == data_shape:
        input_model.dq[is_invalid] |= dqflags.pixel['DO_NOT_USE']


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/config.py
"""

"""
Implementation of Step methods related to ASDF config files.  This module
will eventually fully replace config_parser.py, but we'll need to maintain
both until we replace configobj with traitlets.
"""

from copy import deepcopy
from datetime import datetime, timezone

import asdf

from spike.stpipe.utilities import get_fully_qualified_class_name

_CONFIG_SCHEMA_URI = "http://stsci.edu/schemas/stpipe/step_config-1.0.0"
_LEGACY_CONFIG_SCHEMA_URI = "http://stsci.edu/schemas/stpipe/step_config-0.1.0"

# The config schema validates that this substring is not present:
_TEMPLATE_PLACEHOLDER = "<SPECIFY>"

_META_TEMPLATE = {
    "author": _TEMPLATE_PLACEHOLDER,
    "date": _TEMPLATE_PLACEHOLDER,
    "description": f"Parameters for calibration step {_TEMPLATE_PLACEHOLDER}",
    "instrument": {
        "name": _TEMPLATE_PLACEHOLDER,
    },
    "origin": _TEMPLATE_PLACEHOLDER,
    "pedigree": _TEMPLATE_PLACEHOLDER,
    "reftype": _TEMPLATE_PLACEHOLDER,
    "telescope": _TEMPLATE_PLACEHOLDER,
    "useafter": _TEMPLATE_PLACEHOLDER,
}


class StepConfig:
    """
    Step configuration container.

    Parameters
    ----------
    class_name : str
        Fully-qualified Step subclass name.
    name : str
        Nickname
    parameters : dict
        Parameters indexed by parameter name.
    steps : list of StepConfig
        List of sub-step configs.
    """

    def __init__(self, class_name, name, parameters, steps):
        self._class_name = class_name
        self._name = name
        self._parameters = parameters
        self._steps = steps

    @property
    def class_name(self):
        return self._class_name

    @property
    def name(self):
        return self._name

    @property
    def parameters(self):
        return self._parameters

    @property
    def steps(self):
        return self._steps

    def __eq__(self, other):
        if not isinstance(other, StepConfig):
            return False

        return (
            self.class_name == other.class_name
            and self.name == other.name
            and self.parameters == other.parameters
            and self.steps == other.steps
        )

    def _to_tree(self):
        tree = {
            "class": self.class_name,
            "name": self.name,
            "parameters": self.parameters,
        }
        if len(self.steps) > 0:
            tree["steps"] = [s._to_tree() for s in self.steps]
        return tree

    def to_asdf(self, include_metadata=False):
        """
        Convert this config to an AsdfFile, which may be
        used to write the config to a file.

        Parameters
        ----------
        include_metadata : bool, optional
            Set to True to include metadata that is required
            for submission to CRDS.

        Returns
        -------
        asdf.AsdfFile
        """
        result = asdf.AsdfFile(self._to_tree())

        if include_metadata:
            meta = deepcopy(_META_TEMPLATE)
            meta["date"] = meta["date"].replace(
                _TEMPLATE_PLACEHOLDER,
                datetime.now(timezone.utc)
                .replace(microsecond=0)
                .isoformat()
                .removesuffix("+00:00"),
            )
            meta["description"] = meta["description"].replace(
                _TEMPLATE_PLACEHOLDER, self.class_name
            )
            result["meta"] = meta

        _validate_asdf(result, _CONFIG_SCHEMA_URI)

        return result

    @classmethod
    def from_asdf(cls, asdf_file):
        """
        Create a StepConfig from an open AsdfFile.

        Parameters
        ----------
        asdf_file : asdf.AsdfFile

        Returns
        -------
        stpipe.config.StepConfig

        Raises
        ------
        asdf.ValidationError
            If the file does not validate against any of the
            recognized schemas.
        """
        try:
            _validate_asdf(asdf_file, _CONFIG_SCHEMA_URI)
            return cls._from_tree(asdf_file.tree)
        except asdf.ValidationError as e:
            # Maybe it's an old-style ASDF file:
            try:
                _validate_asdf(asdf_file, _LEGACY_CONFIG_SCHEMA_URI)
                return cls._from_legacy_tree(asdf_file.tree)
            except asdf.ValidationError as err:
                # Raise the original error so that we encourage
                # use of the new config file format.
                raise e from err

    @classmethod
    def _from_tree(cls, tree):
        class_name = tree["class"]
        name = tree["name"]
        parameters = deepcopy(tree["parameters"])
        steps = [cls._from_tree(s) for s in tree.get("steps", [])]
        return StepConfig(class_name, name, parameters, steps)

    @classmethod
    def _from_legacy_tree(cls, tree):
        def _inner(parameters, name):
            class_name = parameters.pop("class", None)
            steps = [_inner(s, n) for n, s in parameters.pop("steps", {}).items()]
            return StepConfig(class_name, name, parameters, steps)

        parameters = deepcopy(tree["parameters"])
        name = parameters.pop("name")
        return _inner(parameters, name)


def export_config(step):
    """
    Export a step's current parameters to a StepConfig object.

    Parameters
    ----------
    step : stpipe.Step

    Returns
    -------
    stpipe.config.StepConfig
    """
    class_name = get_fully_qualified_class_name(step)
    name = step.name

    parameters = step.get_pars()
    # The Pipeline class includes step parameters, but we're
    # going to collect those ourselves so we can also get
    # ahold of the name and class name.
    parameters.pop("steps", None)

    steps = [
        export_config(getattr(step, step_name))
        for step_name, _ in getattr(step, "step_defs", {}).items()
    ]

    return StepConfig(class_name, name, parameters, steps)


def _validate_asdf(asdf_file, schema_uri):
    """
    Validate an ASDF file against the config schema.

    Parameters
    ----------
    asdf_file : asdf.AsdfFile
        Open config file.
    schema_uri : str
        StepConfig schema URI (must be registered with the ASDF entry point).

    Raises
    ------
    asdf.ValidationError
    """
    # TODO: We should add a method on AsdfFile to facilitate this:
    schema = asdf.schema.load_schema(schema_uri)
    tagged_tree = asdf.yamlutil.custom_tree_to_tagged_tree(asdf_file.tree, asdf_file)
    asdf.schema.validate(tagged_tree, asdf_file, schema)


import pytest
from astropy.coordinates import SkyCoord
import astropy.units as u
from paralleltest import paralleltest #local file and function
from spike.tools import objloc

@pytest.mark.parametrize(
	"obj, expected",
	[
		(objloc('M51'), 
		SkyCoord('202.469575 47.19525833', unit = (u.deg, u.deg), frame = 'icrs')),

		(objloc('10:00:30.03 +02:08:59.47'), 
		SkyCoord('10:00:30.03 +02:08:59.47', unit = (u.hour, u.deg), frame = 'icrs')),

		(objloc('150.125125 2.1498528'),
		SkyCoord('150.125125 2.1498528', unit = (u.deg, u.deg), frame = 'icrs'))
		
	])

def test_objloc(obj, expected):
	from spike.tools import objloc
	assert obj == expected


def test_multiprocessing():
	# simple check on async file creation
	# if this doesn't pass on the assert, user should not use the parallel option
	import os
	from multiprocessing import Pool, cpu_count
	pool = Pool(processes = (cpu_count() - 1))
	for i in ['a', 'b', 'c', 'd', 'e', 'f']:
		pool.apply_async(paralleltest, args = (i))
	pool.close()
	pool.join()
	for i in ['a', 'b', 'c', 'd', 'e', 'f']:
		with open(i+'.txt') as file:
			checkstring = file.read()
		assert i+'\n' == checkstring
		os.system('rm '+i+'_test.txt')






"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/tweakreg/tweakreg_step.py
"""


import os
from pathlib import Path

import numpy as np
from astropy.table import Table
from roman_datamodels import datamodels as rdm
import spike.stcal.tweakreg_tweakreg as tweakreg
from spike.stcal.tweakreg_tweakreg import _SINGLE_GROUP_REFCAT_STR, SINGLE_GROUP_REFCAT

from spike.romancal.assign_wcs_utils import add_s_region

# LOCAL
from spike.romancal.datamodels_library import ModelLibrary
from spike.romancal.stpipe_core import RomanStep

DEFAULT_ABS_REFCAT = SINGLE_GROUP_REFCAT[0]

__all__ = ["TweakRegStep"]


class TweakRegStep(RomanStep):
    """
    TweakRegStep: Image alignment based on catalogs of sources detected in
    input images.
    """

    class_alias = "tweakreg"

    spec = f"""
        use_custom_catalogs = boolean(default=False) # Use custom user-provided catalogs?
        catalog_format = string(default='ascii.ecsv') # Catalog output file format
        catfile = string(default='') # Name of the file with a list of custom user-provided catalogs
        catalog_path = string(default='') # Catalog output file path
        enforce_user_order = boolean(default=False) # Align images in user specified order?
        expand_refcat = boolean(default=False) # Expand reference catalog with new sources?
        minobj = integer(default=15) # Minimum number of objects acceptable for matching
        searchrad = float(default=2.0) # The search radius in arcsec for a match
        use2dhist = boolean(default=True) # Use 2d histogram to find initial offset?
        separation = float(default=1.0) # Minimum object separation in arcsec
        tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec
        fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift') # Fitting geometry
        nclip = integer(min=0, default=3) # Number of clipping iterations in fit
        sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units
        abs_refcat = string(default='{DEFAULT_ABS_REFCAT}')  # Absolute reference
        # catalog. Options: {_SINGLE_GROUP_REFCAT_STR}
        save_abs_catalog = boolean(default=False)  # Write out used absolute astrometric reference catalog as a separate product
        abs_minobj = integer(default=15) # Minimum number of objects acceptable for matching when performing absolute astrometry
        abs_searchrad = float(default=6.0) # The search radius in arcsec for a match when performing absolute astrometry
        # We encourage setting this parameter to True. Otherwise, xoffset and yoffset will be set to zero.
        abs_use2dhist = boolean(default=True) # Use 2D histogram to find initial offset when performing absolute astrometry?
        abs_separation = float(default=1.0) # Minimum object separation in arcsec when performing absolute astrometry
        abs_tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec when performing absolute astrometry
        # Fitting geometry when performing absolute astrometry
        abs_fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift')
        abs_nclip = integer(min=0, default=3) # Number of clipping iterations in fit when performing absolute astrometry
        abs_sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units when performing absolute astrometry
        output_use_model = boolean(default=True)  # When saving use `DataModel.meta.filename`
        update_source_catalog_coordinates = boolean(default=False) # Update source catalog file with tweaked coordinates?
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):

        # properly handle input
        try:
            if isinstance(input, rdm.DataModel):
                images = ModelLibrary([input])
            elif str(input).endswith(".asdf"):
                images = ModelLibrary([rdm.open(input)])
            elif isinstance(input, ModelLibrary):
                images = input
            else:
                images = ModelLibrary(input)
        except TypeError as e:
            e.args = (
                "Input to tweakreg must be a list of DataModels, an "
                "association, or an already open ModelLibrary "
                "containing one or more DataModels.",
            ) + e.args[1:]
            raise e

        if not images:
            raise ValueError("Input must contain at least one image model.")

        self.log.info(
            f"Number of image groups to be aligned: {len(images.group_indices):d}."
        )
        self.log.info("Image groups:")
        for name in images.group_names:
            self.log.info(f"  {name}")
        # set the first image as reference
        with images:
            ref_image = images.borrow(0)
            images.shelve(ref_image, 0, modify=False)

        catdict = _parse_catfile(self.catfile)

        use_custom_catalogs = self.use_custom_catalogs
        # if user requested the use of custom catalogs and provided a
        # valid 'catfile' file name that has no custom catalogs,
        # turn off the use of custom catalogs:
        if catdict is not None and not catdict:
            self.log.warning(
                "'use_custom_catalogs' is set to True but 'catfile' "
                "contains no user catalogs."
            )
            use_custom_catalogs = False

        if use_custom_catalogs and catdict:
            with images:
                for i, member in enumerate(images.asn["products"][0]["members"]):
                    filename = member["expname"]
                    if filename in catdict:
                        # FIXME: I'm not sure if this captures all the possible combinations
                        # for example, meta.tweakreg_catalog is set by the container (when
                        # it's present in the association). However the code in this step
                        # checks meta.source_catalog.tweakreg_catalog. I think this means
                        # that setting a catalog via an association does not work. Is this
                        # intended? If so, the container can be updated to not support that.
                        model = images.borrow(i)
                        model.meta["source_catalog"] = {
                            "tweakreg_catalog_name": catdict[filename],
                        }
                        images.shelve(model, i)
                    else:
                        images.shelve(model, i, modify=False)

        # set path where the source catalog will be saved to
        if len(self.catalog_path) == 0:
            self.catalog_path = os.getcwd()
        self.catalog_path = Path(self.catalog_path).as_posix()
        self.log.info(f"All source catalogs will be saved to: {self.catalog_path}")

        # set reference catalog name
        if not self.abs_refcat:
            self.abs_refcat = DEFAULT_ABS_REFCAT.strip().upper()
        if self.abs_refcat != DEFAULT_ABS_REFCAT:
            self.expand_refcat = True

        # build the catalogs for input images
        imcats = []
        with images:
            for i, image_model in enumerate(images):
                exposure_type = image_model.meta.exposure.type
                if exposure_type != "WFI_IMAGE":
                    self.log.info("Skipping TweakReg for spectral exposure.")
                    image_model.meta.cal_step.tweakreg = "SKIPPED"
                else:
                    source_catalog = getattr(image_model.meta, "source_catalog", None)
                    if source_catalog is None:
                        images.shelve(image_model, i, modify=False)
                        raise AttributeError(
                            "Attribute 'meta.source_catalog' is missing. "
                            "Please either run SourceCatalogStep or provide a custom source catalog."
                        )

                    try:
                        catalog = self.get_tweakreg_catalog(source_catalog, image_model)
                    except AttributeError as e:
                        self.log.error(f"Failed to retrieve tweakreg_catalog: {e}")
                        images.shelve(image_model, i, modify=False)
                        raise e

                    try:
                        # validate catalog columns
                        _validate_catalog_columns(catalog)
                    except ValueError as e:
                        self.log.error(f"Failed to validate catalog columns: {e}")
                        images.shelve(image_model, i, modify=False)
                        raise e

                    catalog = tweakreg.filter_catalog_by_bounding_box(
                        catalog, image_model.meta.wcs.bounding_box
                    )

                    if self.save_abs_catalog:
                        output_name = os.path.join(
                            self.catalog_path, f"fit_{self.abs_refcat.lower()}_ref.ecsv"
                        )
                        catalog.write(
                            output_name, format=self.catalog_format, overwrite=True
                        )

                    image_model.meta["tweakreg_catalog"] = catalog.as_array()
                    nsources = len(catalog)
                    self.log.info(
                        f"Detected {nsources} sources in {image_model.meta.filename}."
                        if nsources
                        else f"No sources found in {image_model.meta.filename}."
                    )
                    # build image catalog
                    # catalog name
                    catalog_name = os.path.splitext(image_model.meta.filename)[0].strip(
                        "_- "
                    )
                    # catalog data
                    catalog_table = Table(image_model.meta.tweakreg_catalog)
                    catalog_table.meta["name"] = catalog_name

                    imcat = tweakreg.construct_wcs_corrector(
                        wcs=image_model.meta.wcs,
                        refang=image_model.meta.wcsinfo,
                        catalog=catalog_table,
                        group_id=image_model.meta.group_id,
                    )
                    imcat.meta["model_index"] = i
                    imcats.append(imcat)
                images.shelve(image_model, i)

        # run alignment only if it was possible to build image catalogs
        if len(imcats):
            # extract WCS correctors to use for image alignment
            if len(images.group_indices) > 1:
                self.do_relative_alignment(imcats)

            if self.abs_refcat in SINGLE_GROUP_REFCAT:
                self.do_absolute_alignment(ref_image, imcats)

            # finalize step
            with images:
                for imcat in imcats:
                    image_model = images.borrow(imcat.meta["model_index"])
                    image_model.meta.cal_step["tweakreg"] = "COMPLETE"
                    # remove source catalog
                    del image_model.meta["tweakreg_catalog"]

                    # retrieve fit status and update wcs if fit is successful:
                    if "SUCCESS" in imcat.meta.get("fit_info")["status"]:
                        # Update/create the WCS .name attribute with information
                        # on this astrometric fit as the only record that it was
                        # successful:

                        # NOTE: This .name attrib agreed upon by the JWST Cal
                        #       Working Group.
                        #       Current value is merely a place-holder based
                        #       on HST conventions. This value should also be
                        #       translated to the FITS WCSNAME keyword
                        #       IF that is what gets recorded in the archive
                        #       for end-user searches.
                        imcat.wcs.name = f"FIT-LVL2-{self.abs_refcat}"

                        # serialize object from tweakwcs
                        # (typecasting numpy objects to python types so that it doesn't cause an
                        # issue when saving datamodel to ASDF)
                        wcs_fit_results = {
                            k: (
                                v.tolist()
                                if isinstance(v, (np.ndarray, np.bool_))
                                else v
                            )
                            for k, v in imcat.meta["fit_info"].items()
                        }
                        # add fit results and new WCS to datamodel
                        image_model.meta["wcs_fit_results"] = wcs_fit_results
                        # remove unwanted keys from WCS fit results
                        for k in [
                            "eff_minobj",
                            "matched_ref_idx",
                            "matched_input_idx",
                            "fit_RA",
                            "fit_DEC",
                            "fitmask",
                        ]:
                            del image_model.meta["wcs_fit_results"][k]

                        # update WCS
                        image_model.meta.wcs = imcat.wcs
                        # update S_REGION
                        add_s_region(image_model)

                    images.shelve(image_model, imcat.meta["model_index"])

        return images

    def update_catalog_coordinates(self, tweakreg_catalog_name, tweaked_wcs):
        """
        Update the source catalog coordinates using the tweaked WCS.

        Parameters
        ----------
        tweakreg_catalog_name : str
            The name of the TweakReg catalog file produced by `SourceCatalog`.
        tweaked_wcs : `gwcs.wcs.WCS`
            The tweaked World Coordinate System (WCS) object.

        Returns
        -------
        None
        """
        # read in cat file
        with rdm.open(tweakreg_catalog_name) as source_catalog_model:
            # get catalog
            catalog = source_catalog_model.source_catalog

            # define mapping between pixel and world coordinates
            colname_mapping = {
                ("xcentroid", "ycentroid"): ("ra_centroid", "dec_centroid"),
                ("x_psf", "y_psf"): ("ra_psf", "dec_psf"),
            }

            for k, v in colname_mapping.items():
                # get column names
                x_colname, y_colname = k
                ra_colname, dec_colname = v

                # calculate new coordinates using tweaked WCS and update catalog coordinates
                catalog[ra_colname], catalog[dec_colname] = tweaked_wcs(
                    catalog[x_colname], catalog[y_colname]
                )

            # save updated catalog (overwrite cat file)
            self.save_model(
                source_catalog_model,
                output_file=source_catalog_model.meta.filename,
                suffix="cat",
                force=True,
            )

    def read_catalog(self, catalog_name):
        """
        Reads a source catalog from a specified file.

        This function determines the format of the catalog based on the file extension.
        If the file ends with "asdf", it uses a specific method to open and read the catalog;
        otherwise, it reads the catalog using a standard table format.

        Parameters
        ----------
        catalog_name : str
            The name of the catalog file to read.

        Returns
        -------
        Table
            The read catalog as a Table object.

        Raises
        ------
        ValueError
            If the catalog format is unsupported.
        """
        if catalog_name.endswith("asdf"):
            with rdm.open(catalog_name) as source_catalog_model:
                catalog = source_catalog_model.source_catalog
        else:
            catalog = Table.read(catalog_name, format=self.catalog_format)
        return catalog

    def get_tweakreg_catalog(self, source_catalog, image_model):
        """
        Retrieve the tweakreg catalog from source detection.

        This method checks the source detection metadata for the presence of a
        tweakreg catalog data or a string with its name. It returns the catalog
        as a Table object if either is found, or raises an error if neither is available.

        Parameters
        ----------
        source_catalog : object
            The source catalog metadata containing catalog information.
        image_model : DataModel
            The image model associated with the source detection.

        Returns
        -------
        Table
            The retrieved tweakreg catalog as a Table object.

        Raises
        ------
        AttributeError
            If the required catalog information is missing from the source detection.
        """
        if getattr(source_catalog, "tweakreg_catalog", None):
            tweakreg_catalog = Table(np.asarray(source_catalog.tweakreg_catalog))
            del image_model.meta.source_catalog["tweakreg_catalog"]
            return tweakreg_catalog

        if getattr(source_catalog, "tweakreg_catalog_name", None):
            return self.read_catalog(source_catalog.tweakreg_catalog_name)

        raise AttributeError(
            "Attribute 'meta.source_catalog.tweakreg_catalog' is missing. "
            "Please either run SourceCatalogStep or provide a custom source catalog."
        )

    def do_relative_alignment(self, imcats):
        """
        Perform relative alignment of images.

        This method performs relative alignment with the specified parameters,
        including search radius, separation, and fitting geometry.

        Parameters
        ----------
        imcats : list
            A list of image catalogs containing source information for alignment.

        Returns
        -------
        None
        """
        tweakreg.relative_align(
            imcats,
            searchrad=self.searchrad,
            separation=self.separation,
            use2dhist=self.use2dhist,
            tolerance=self.tolerance,
            xoffset=0,
            yoffset=0,
            enforce_user_order=self.enforce_user_order,
            expand_refcat=self.expand_refcat,
            minobj=self.minobj,
            fitgeometry=self.fitgeometry,
            nclip=self.nclip,
            sigma=self.sigma,
            clip_accum=True,
        )

    def do_absolute_alignment(self, ref_image, imcats):
        """
        Perform absolute alignment of images.

        This method retrieves a reference image and performs absolute alignment
        using the specified parameters, including reference WCS information and
        catalog details. It aligns the provided image catalogs to the absolute
        reference catalog.

        Parameters
        ----------
        ref_image : DataModel
            The reference image used for alignment, which contains WCS information.
        imcats : list
            A list of image catalogs containing source information for alignment.

        Returns
        -------
        None
        """
        tweakreg.absolute_align(
            imcats,
            self.abs_refcat,
            ref_wcs=ref_image.meta.wcs,
            ref_wcsinfo=ref_image.meta.wcsinfo,
            epoch=ref_image.meta.exposure.mid_time.decimalyear,
            abs_minobj=self.abs_minobj,
            abs_fitgeometry=self.abs_fitgeometry,
            abs_nclip=self.abs_nclip,
            abs_sigma=self.abs_sigma,
            abs_searchrad=self.abs_searchrad,
            abs_use2dhist=self.abs_use2dhist,
            abs_separation=self.abs_separation,
            abs_tolerance=self.abs_tolerance,
            save_abs_catalog=self.save_abs_catalog,
            abs_catalog_output_dir=self.output_dir,
            clip_accum=True,
        )


def _parse_catfile(catfile):
    """
    Parse a catalog file and return a dictionary mapping data models to catalog paths.

    This function reads a specified catalog file, extracting data model names and
    their associated catalog paths. It supports a format where each line contains
    a data model followed by an optional catalog path, and it ensures that the
    file adheres to the expected structure.

    Parameters
    ----------
    catfile : str
        The path to the catalog file to be parsed.

    Returns
    -------
    dict or None
        A dictionary mapping data model names to catalog paths, or None if the
        input file is empty or invalid.

    Raises
    ------
    ValueError
        If the catalog file contains more than two columns per line.
    """

    if catfile is None or not catfile.strip():
        return None

    catdict = {}

    with open(catfile) as f:
        catfile_dir = os.path.dirname(catfile)

        for line in f:
            sline = line.strip()
            if not sline or sline[0] == "#":
                continue

            data_model, *catalog = sline.split()
            catalog = list(map(str.strip, catalog))
            if len(catalog) == 1:
                catdict[data_model] = os.path.join(catfile_dir, catalog[0])
            elif not catalog:
                catdict[data_model] = None
            else:
                raise ValueError("'catfile' can contain at most two columns.")

    return catdict


def _validate_catalog_columns(catalog):
    """
    Validate the presence of required columns in the catalog.

    This method checks if the specified axis column exists in the catalog.
    If the axis is not found, it looks for a corresponding centroid column
    and renames it if present. If neither is found, it raises an error.

    Parameters
    ----------
    catalog : Table
        The catalog to validate, which should contain source information.
    axis : str
        The axis to check for in the catalog (e.g., 'x' or 'y').

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the required columns are missing from the catalog.
    """
    for axis in ["x", "y"]:
        if axis not in catalog.colnames:
            long_axis = f"{axis}centroid"
            if long_axis in catalog.colnames:
                catalog.rename_column(long_axis, axis)
            else:
                raise ValueError(
                    "'tweakreg' source catalogs must contain a header with "
                    "columns named either 'x' and 'y' or 'xcentroid' and 'ycentroid'."
                )
    return catalog



def libpath(filepath):
    '''Return the full path to the module library.'''
    from os.path import (
        abspath,
        dirname,
        join
    )
    return join(dirname(abspath(__file__)),
                'lib',
                filepath)


__all__ = ['resample_step', 'resample', 'resample_utils', 
'assign_wcs_util', 'model_blender_blender', 'associations_lib_roles_level3_base', 
'associations_lib_acid', 'associations_lib_dms_base', 
'associations_lib_constraint', 'associations_association', 
'associations_lib_ioregistry', 'associations_lib_keyvalue_registry.py',
'associations_registry','associations_lib_callback_registry',
'associations_lib_diff', 'associations_lib_prune',
'lib_catalog_utils', 'lib_suffix', 'lib_signal_slot'
'associations_lib_prune', 'associations_config', 'associations_lib_product_utils',
'associations_load_asn' 'associations_asn_from_list',
'associations_exceptions', 'associations_lib_process_list', 'associations_lib_utilities', 
'associations_pool', 'associations_lib_counter', 'datamodels_library', 'model_blender_rules', 
'model_blender_schemautil', 'model_blender_tablebuilder', 'stpipe_utilities',
'tweakreg_tweakreg_step', 'stpipe_core',
'tweakreg_tweakreg_catalog', 'source_catalog_detection', 'lib_pipe_utils']


from . import stpipe_utilities
from . import resample_step, resample, resample_utils
from . import assign_wcs_util, model_blender_blender
from . import associations_lib_rules_level3_base, associations_lib_acid
from . import associations_lib_constraint, associations_association, associations_lib_dms_base
from . import associations_lib_keyvalue_registry, associations_lib_ioregistry
from . import associations_registry, associations_lib_callback_registry
from . import associations_lib_prune, associations_lib_diff
from . import lib_catalog_utils, lib_suffix, lib_signal_slot
from . import associations_lib_member, associations_config, associations_lib_product_utils
from . import associations_load_asn, associations_asn_from_list
from . import associations_exceptions, associations_lib_process_list, associations_lib_utilities
from . import associations_pool, associations_lib_counter, datamodels_library, model_blender_rules
from . import model_blender_schemautil, model_blender_tablebuilder
from . import tweakreg_tweakreg_step, stpipe_core
from . import tweakreg_tweakreg_catalog, source_catalog_detection
from . import lib_pipe_utils


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/lib/rules_level3_base.py
"""

from collections import defaultdict
import logging
from os.path import (
    basename,
    split,
    splitext
    )
import re

# from jwst.associations import (
#     Association,
#     ListCategory,
#     libpath
# )
from spike.jwstcal.associations_association import Association
from spike.jwstcal.associations_lib_process_list import ListCategory
from spike.jwstcal import libpath

from spike.jwstcal.associations_registry import RegistryMarker
from spike.jwstcal.associations_lib_utilities import (
    evaluate,
    is_iterable
)
from spike.jwstcal.associations_exceptions import (
    AssociationNotValidError,
)
from spike.jwstcal.associations_lib_acid import ACID
from spike.jwstcal.associations_lib_constraint import (
    Constraint,
    SimpleConstraint,
)
from spike.jwstcal.associations_lib_counter import Counter
from spike.jwstcal.associations_lib_dms_base import (
    _EMPTY,
    Constraint_TargetAcq,
    DMSAttrConstraint,
    DMSBaseMixin,
    IMAGE2_SCIENCE_EXP_TYPES,
    IMAGE2_NONSCIENCE_EXP_TYPES,
    SPEC2_SCIENCE_EXP_TYPES,
)
from spike.stpipe.format_template import FormatTemplate
from spike.jwstcal.associations_lib_member import Member
from spike.jwstcal.associations_lib_prune import prune



__all__ = [
    'ASN_SCHEMA',
    'AsnMixin_AuxData',
    'AsnMixin_Coronagraphy',
    'AsnMixin_Science',
    'AsnMixin_Spectrum',
    'Constraint',
    'Constraint_Base',
    'Constraint_IFU',
    'Constraint_Image',
    'Constraint_MSA',
    'Constraint_Obsnum',
    'Constraint_Optical_Path',
    'Constraint_Spectral',
    'Constraint_Target',
    'DMS_Level3_Base',
    'DMSAttrConstraint',
    'ListCategory',
    'SimpleConstraint',
    'Utility',
]
from spike.jwstcal.lib_suffix import remove_suffix
# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# The schema that these associations must adhere to.
ASN_SCHEMA = RegistryMarker.schema(libpath('asn_schema_jw_level3.json'))

# DMS file name templates
_LEVEL1B_REGEX = r'(?P<path>.+)(?P<type>_uncal)(?P<extension>\..+)'
_DMS_POOLNAME_REGEX = r'jw(\d{5})_(\d{8}[Tt]\d{6})_pool'

# Product name regex's
_REGEX_ACID_VALUE = r'(o\d{3}|(c|a)\d{4})'

# Exposures that should have received Level2b processing
LEVEL2B_EXPTYPES = []
LEVEL2B_EXPTYPES.extend(IMAGE2_SCIENCE_EXP_TYPES)
LEVEL2B_EXPTYPES.extend(IMAGE2_NONSCIENCE_EXP_TYPES)
LEVEL2B_EXPTYPES.extend(SPEC2_SCIENCE_EXP_TYPES)

# Association Candidates that should have more than one observations
MULTI_OBS_AC_TYPES = ['group', 'background', 'mosaic']


class DMS_Level3_Base(DMSBaseMixin, Association):
    """Basic class for DMS Level3 associations."""

    # Set the validation schema
    schema_file = ASN_SCHEMA.schema

    # Attribute values that are indicate the
    # attribute is not specified.
    INVALID_VALUES = _EMPTY

    # Make sequences type-dependent
    _sequences: defaultdict = defaultdict(Counter)

    def __init__(self, *args, **kwargs):

        super(DMS_Level3_Base, self).__init__(*args, **kwargs)

        # Initialize validity checks
        self.validity.update({
            'has_science': {
                'validated': False,
                'check': lambda member: member['exptype'] == 'science'
            },
        })

        # Other presumptions on the association
        if 'constraints' not in self.data:
            self.data['constraints'] = 'No constraints'
        if 'asn_type' not in self.data:
            self.data['asn_type'] = 'user_built'
        if 'asn_id' not in self.data:
            self.data['asn_id'] = 'a3001'
        if 'target' not in self.data:
            self.data['target'] = 'none'
        if 'asn_pool' not in self.data:
            self.data['asn_pool'] = 'none'

    @property
    def current_product(self):
        return self.data['products'][-1]

    @property
    def dms_product_name(self):
        """Define product name.

        Returns
        -------
        product_name : str
            The product name
        """
        return self._dms_product_name(self)

    @staticmethod
    def _dms_product_name(association):
        """Define product name.

        Parameters
        ----------
        association : `Association`
            Association to get the name from.

        Returns
        -------
        product_name : str
            The product name
        """
        target = association._get_target()

        instrument = association._get_instrument()

        opt_elem = association._get_opt_element()

        slit_name = association._get_slit_name()
        if slit_name:
            slit_name = '-' + slit_name

        exposure = association._get_exposure()
        if exposure:
            exposure = '-' + exposure

        subarray = association._get_subarray()
        if subarray:
            subarray = '-' + subarray

        product_name = (
            'jw{program}-{acid}'
            '_{target}'
            '_{instrument}'
            '_{opt_elem}{slit_name}{subarray}'
        )
        product_name = product_name.format(
            program=association.data['program'],
            acid=association.acid.id,
            target=target,
            instrument=instrument,
            opt_elem=opt_elem,
            slit_name=slit_name,
            subarray=subarray,
            exposure=exposure
        )

        return product_name.lower()

    def update_asn(self, item=None, member=None):
        """Update association meta information

        Parameters
        ----------
        item : dict or None
            Item to use as a source. If not given, item-specific
            information will be left unchanged.

        member : Member or None
            An association member to use as source.
            If not given, member-specific information will be update
            from current association/product membership.

        Notes
        -----
        If both `item` and `member` are given,
        information in `member` will take precedence.
        """
        super(DMS_Level3_Base, self).update_asn(item=item, member=member)

        # Constraints
        self.data['constraints'] = str(self.constraints)

        # ID
        self.data['asn_id'] = self.acid.id

        # Target
        self.data['target'] = self._get_target()

        # Item-based information
        if item is not None:

            # Program
            if self.data['program'] == 'noprogram':
                self.data['program'] = '{:0>5s}'.format(item['program'])

            # Pool
            if self.data['asn_pool'] == 'none':
                self.data['asn_pool'] = basename(
                    item.meta['pool_file']
                )
                parsed_name = re.search(
                    _DMS_POOLNAME_REGEX, self.data['asn_pool'].split('.')[0]
                )
                if parsed_name is not None:
                    pool_meta = {
                        'program_id': parsed_name.group(1),
                        'version': parsed_name.group(2)
                    }
                    self.meta['pool_meta'] = pool_meta

        # Product-based updates
        product = self.current_product
        product['name'] = self.dms_product_name

    def make_member(self, item):
        """Create a member from the item

        Parameters
        ----------
        item : dict
            The item to create member from.

        Returns
        -------
        member : Member
            The member
        """
        try:
            exposerr = item['exposerr']
        except KeyError:
            exposerr = None

        # Get exposure type
        exptype = self.get_exposure_type(item)

        # Determine expected member name
        expname = Utility.rename_to_level2(
            item['filename'],
            exp_type=item['exp_type'],
            use_integrations=(self.is_item_coron(item) |
                              # NIS_AMI used to use cal files;
                              # now switched to calints
                              self.is_item_ami(item) |
                              self.is_item_tso(item)),
            member_exptype=exptype
        )

        member = Member(
            {
                'expname': expname,
                'exptype': exptype,
                'exposerr': exposerr,
                'asn_candidate': item['asn_candidate']
            },
            item=item
        )
        return member

    def make_fixedslit_bkg(self):
        """Add a background to a MIR_lrs-fixedslit observation"""

        # check to see if these are nodded backgrounds, if they are setup
        # the background members, otherwise return the original association
        # to test for the string 'nod' we need to copy and pop the value out of the set
        if 'nod' not in self.constraints['patttype_spectarg'].found_values.copy().pop():
            results = []
            results.append(self)
            return results

        for product in self['products']:
            members = product['members']
            # Split out the science exposures
            science_exps = [
                member
                for member in members
                if member['exptype'] == 'science'
            ]
            # if there is only one science observation it cannot be the background
            # return with original association.
            if len(science_exps) < 2:
                return results

            # Create new members for each science exposure in the association,
            # using the the base name + _x1d as background.
            results = []
            # Loop over all science exposures in the association
            for science_exp in science_exps:
                sci_name = science_exp['expname']
                science_exp['expname'] = sci_name
                # Construct the name for the background file
                bkg_name = remove_suffix(
                    splitext(split(science_exp['expname'])[1])[0])[0]
                bkg_name = bkg_name+'_x1d.fits'
                now_background = Member(science_exp)
                now_background['expname'] = bkg_name
                now_background['exptype'] = 'background'
                # Add the background file to the association table
                members.append(now_background)

            if self.is_valid:
                results.append(self)

            return results


    def _init_hook(self, item):
        """Post-check and pre-add initialization"""
        super(DMS_Level3_Base, self)._init_hook(item)

        # Set which sequence counter should be used.
        self._sequence = self._sequences[self.data['asn_type']]

        # Create the product.
        self.new_product()

        # Update meta data
        self.update_asn(item=item)

    def _add(self, item):
        """Add item to this association."""
        member = self.make_member(item)
        if self.is_member(member):
            return

        self.update_validity(member)
        members = self.current_product['members']
        members.append(member)
        if member['exposerr'] not in _EMPTY:
            logger.warning('Member {} has exposure error "{}"'.format(
                item['filename'],
                member['exposerr']
            ))

        # Update meta info
        self.update_asn(item=item, member=member)

    def _add_items(self,
                   items,
                   product_name=None,
                   with_exptype=False,
                   **kwargs):
        """ Force adding items to the association

        Parameters
        ----------
        items : [object[, ...]]
            A list of items to make members of the association.

        product_name : str or None
            The name of the product to add the items to.
            If the product does not already exist, it will be created.
            If None, the default DMS Level3 naming
            conventions will be attempted.

        with_exptype : bool
            If True, each item is expected to be a 2-tuple with
            the first element being the item to add as `expname`
            and the second items is the `exptype`

        kwargs : dict
            Allows other keyword arguments used by other subclasses.

        Notes
        -----
        This is a low-level shortcut into adding members, such as file names,
        to an association. All defined shortcuts and other initializations are
        by-passed, resulting in a potentially unusable association.
        """
        if product_name is None:
            raise AssociationNotValidError(
                'Product name needs to be specified'
            )

        self.new_product(product_name)
        members = self.current_product['members']
        for item in items:
            exptype = 'science'
            if with_exptype:
                item, exptype = item
            member = Member(
                {
                    'expname': item,
                    'exptype': exptype
                },
                item=item
            )
            self.update_validity(member)
            members.append(member)
        self.sequence = next(self._sequence)

    def __repr__(self):
        try:
            file_name, json_repr = self.ioregistry['json'].dump(self)
        except:
            return str(self.__class__)
        return json_repr

    def __str__(self):
        result_list = []
        result_list.append(
            '{} with {} products'.format(
                self.asn_name,
                len(self.data['products'])
            )
        )
        result_list.append(
            'Rule={}'.format(self.data['asn_rule'])
        )
        result_list.append(self.data['constraints'])
        result_list.append('Products:')
        for product in self.data['products']:
            result_list.append(
                '\t{} with {} members'.format(
                    product['name'],
                    len(product['members'])
                )
            )
        result = '\n'.join(result_list)
        return result


@RegistryMarker.utility
class Utility():
    """Utility functions that understand DMS Level 3 associations"""

    @staticmethod
    def resequence(associations):
        """Resequence the numbering for the Level3 association types"""
        counters = defaultdict(lambda: defaultdict(Counter))
        for asn in associations:
            asn.sequence = next(
                counters[asn.data['asn_id']][asn.data['asn_type']]
            )

    @staticmethod
    def rename_to_level2(level1b_name, exp_type=None, use_integrations=False, member_exptype='science'):
        """Rename a Level 1b Exposure to a Level2 name.

        The basic transform is changing the suffix `uncal` to
        `cal`, `calints`, or `rate`.

        Parameters
        ----------
        level1b_name : str
            The Level 1b exposure name.

        exp_type:
            JWST exposure type. If not specified,
            it will be presumed that the name
            should get a Level2b name

        use_integrations : boolean
            Use 'calints' instead of 'cal' as the suffix.

        member_exptype: str
            The association member exposure type, such as "science".

        Returns
        -------
        str
            The Level 2b name
        """
        match = re.match(_LEVEL1B_REGEX, level1b_name)
        if match is None or match.group('type') != '_uncal':
            logger.warning((
                'Item FILENAME="{}" is not a Level 1b name. '
                'Cannot transform to Level 2b.'
            ).format(
                level1b_name
            ))
            return level1b_name

        if member_exptype == 'background':
            suffix = 'x1d'
        else:
            if exp_type in LEVEL2B_EXPTYPES:
                suffix = 'cal'
            else:
                suffix = 'rate'

        if use_integrations:
            suffix += 'ints'

        level2_name = ''.join([
            match.group('path'),
            '_',
            suffix,
            match.group('extension')
        ])
        return level2_name

    @staticmethod
    def get_candidate_list(value):
        """Parse the candidate list from a item value

        Parameters
        ----------
        value : str
            The value from the item to parse. Usually
            item['ASN_CANDIDATE']

        Returns
        -------
        [ACID, ...]
            The list of parsed candidates.
        """
        result = []
        evaled = evaluate(value)
        if is_iterable(evaled):
            result = [
                ACID(v)
                for v in evaled
            ]
        return result

    @staticmethod
    @RegistryMarker.callback('finalize')
    def finalize(associations):
        """Check validity and duplications in an association list

        Parameters
        ----------
        associations:[association[, ...]]
            List of associations

        Returns
        -------
        finalized_associations : [association[, ...]]
            The validated list of associations
        """
        finalized_asns = []
        lv3_asns = []
        for asn in associations:
            if isinstance(asn, DMS_Level3_Base):
                finalized = asn.finalize()
                if finalized is not None:
                    lv3_asns.extend(finalized)
            else:
                finalized_asns.append(asn)

        lv3_asns = prune(lv3_asns)

        # Ensure sequencing is correct.
        Utility.resequence(lv3_asns)

        # Merge lists and return
        return finalized_asns + lv3_asns


# ---------
# Utilities
# ---------
# Define default product name filling
format_product = FormatTemplate(
    key_formats={
        'source_id': ['{:s}'],
        'expspcin': ['{:0>2s}'],
        'slit_name': ['{:s}']
    }
)


def dms_product_name_noopt(asn):
    """Define product name without any optical elements.

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """
    target = asn._get_target()

    instrument = asn._get_instrument()

    product_name = 'jw{}-{}_{}_{}'.format(
        asn.data['program'],
        asn.acid.id,
        target,
        instrument
    )
    return product_name.lower()


def dms_product_name_sources(asn):
    """Produce source-based product names

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """
    instrument = asn._get_instrument()

    opt_elem = asn._get_opt_element()

    slit_name = asn._get_slit_name()
    if slit_name:
        slit_name = '-' + slit_name

    subarray = asn._get_subarray()
    if subarray:
        subarray = '-' + subarray

    product_name_format = (
        'jw{program}-{acid}'
        '_{source_id}'
        '_{instrument}'
        '_{opt_elem}{slit_name}{subarray}'
    )
    product_name = format_product(
        product_name_format,
        program=asn.data['program'],
        acid=asn.acid.id,
        instrument=instrument,
        opt_elem=opt_elem,
        slit_name=slit_name,
        subarray=subarray,
    )

    return product_name.lower()


def dms_product_name_nrsfs_sources(asn):
    """Produce source-based product names for NIRSpec fixed-slit observations.

    For this mode, the product names have a placeholder for the
    slit name, to be filled in later by the pipeline.

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """
    instrument = asn._get_instrument()

    opt_elem = asn._get_opt_element()

    subarray = asn._get_subarray()
    if subarray:
        subarray = '-' + subarray

    product_name_format = (
        'jw{program}-{acid}'
        '_{source_id}'
        '_{instrument}'
        '_{opt_elem}-{slit_name}{subarray}'
    )
    product_name = format_product(
        product_name_format,
        program=asn.data['program'],
        acid=asn.acid.id,
        instrument=instrument,
        opt_elem=opt_elem,
        subarray=subarray,
    )

    return product_name.lower()


def dms_product_name_coronimage(asn):
    """Produce image-based product name
       for coronagraphic data

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """

    target = asn._get_target()

    instrument = asn._get_instrument()

    opt_elem = asn._get_opt_element()

    exposure = asn._get_exposure()
    if exposure:
        exposure = '-' + exposure

    subarray = asn._get_subarray()
    if subarray:
        subarray = '-' + subarray

    suffix = '-image3'

    product_name = (
        'jw{program}-{acid}'
        '_{target}'
        '_{instrument}'
        '_{opt_elem}{subarray}{suffix}'
    )
    product_name = product_name.format(
        program=asn.data['program'],
        acid=asn.acid.id,
        target=target,
        instrument=instrument,
        opt_elem=opt_elem,
        subarray=subarray,
        suffix=suffix,
        exposure=exposure
    )

    return product_name.lower()


# -----------------
# Basic constraints
# -----------------
class Constraint_Base(Constraint):
    """Select on program and instrument"""
    def __init__(self):
        super(Constraint_Base, self).__init__(
            [
                DMSAttrConstraint(
                    name='program',
                    sources=['program'],
                ),
                DMSAttrConstraint(
                    name='instrument',
                    sources=['instrume'],
                ),
            ],
            name='base'
        )


class Constraint_IFU(DMSAttrConstraint):
    """Constrain on IFU exposures"""
    def __init__(self):
        super(Constraint_IFU, self).__init__(
            name='exp_type',
            sources=['exp_type'],
            value=(
                'mir_mrs'
                '|mir_flatmrs'
                '|nrs_autowave'
                '|nrs_ifu'
            ),
            force_unique=False
        )


class Constraint_Image(DMSAttrConstraint):
    """Select on exposure type"""
    def __init__(self):
        super(Constraint_Image, self).__init__(
            name='exp_type',
            sources=['exp_type'],
            value=(
                'nrc_image'
                '|mir_image'
                '|nis_image'
                '|fgs_image'
                '|nrs_mimf'
            ),
        )


class Constraint_MSA(Constraint):
    """Constrain on NIRSpec MSA exposures that are spectral"""
    def __init__(self):
        super(Constraint_MSA, self).__init__(
            [
                DMSAttrConstraint(
                    name='exp_type',
                    sources=['exp_type'],
                    value='nrs_msaspec',
                    force_unique=False
                ),
                DMSAttrConstraint(
                    name='is_msa',
                    sources=['msametfl'],
                    force_unique=False,
                )
            ],
            name='msa_spectral'
        )


class Constraint_Obsnum(DMSAttrConstraint):
    """Select on OBSNUM"""
    def __init__(self):
        super(Constraint_Obsnum, self).__init__(
            name='obs_num',
            sources=['obs_num'],
            force_unique=False,
            required=False,
        )


class Constraint_Optical_Path(Constraint):
    """Select on optical path"""
    def __init__(self):
        super(Constraint_Optical_Path, self).__init__([
            DMSAttrConstraint(
                name='opt_elem',
                sources=['filter'],
                required=False,
            ),
            DMSAttrConstraint(
                name='opt_elem2',
                sources=['pupil', 'grating'],
                required=False,
            ),
            Constraint(
                [
                    DMSAttrConstraint(
                        name='fxd_slit2',
                        sources=['fxd_slit'],
                        value=['s200a1|s200a2'],
                        force_unique=False,
                        required=False,
                    ),
                    DMSAttrConstraint(
                        name='fxd_slit',
                        sources=['fxd_slit'],
                        required=False,
                    ),
                ],
                reduce=Constraint.any
            ),
            DMSAttrConstraint(
                name='subarray',
                sources=['subarray']
            )
        ])


class Constraint_Spectral(DMSAttrConstraint):
    """Constrain on spectral exposure types"""
    def __init__(self):
        super(Constraint_Spectral, self).__init__(
            name='exp_type',
            sources=['exp_type'],
            value=(
                'mir_lrs-fixedslit'
                '|nrc_grism'
                '|nrc_wfss'
                '|nrs_autoflat'
                '|nrs_autowave'
                '|nrs_fixedslit'
            ),
            force_unique=False
        )


class Constraint_Target(Constraint):
    """Select on target

    Parameters
    ----------
    association: Association
        If specified, use the `get_exposure_type` method
        to as part of the target selection.
    """
    def __init__(self, association=None):
        constraints = [Constraint([
            DMSAttrConstraint(
                name='acdirect',
                sources=['asn_candidate'],
                value=r"\[\('c\d{4}', 'direct_image'\)\]"
            ),
            SimpleConstraint(
                name='target',
                sources=lambda item: '000'
            )]
        )]
        if association is None:
            constraints.append(DMSAttrConstraint(
                name='target',
                sources=['targetid'],
            ))
        else:
            constraints.append(DMSAttrConstraint(
                name='target',
                sources=['targetid'],
                onlyif=lambda item: association.get_exposure_type(item) != 'background',
                force_reprocess=ListCategory.EXISTING,
                only_on_match=True,
            ))
        super(Constraint_Target, self).__init__(constraints, reduce=Constraint.any)


# -----------
# Base Mixins
# -----------
class AsnMixin_AuxData:
    """Process special and non-science exposures as science.
    """
    def get_exposure_type(self, item, default='science'):
        """Override to force exposure type to always be science
        Parameters
        ----------
        item : dict
            The pool entry for which the exposure type is determined
        default : str or None
            The default exposure type.
            If None, routine will raise LookupError
        Returns
        -------
        exposure_type : 'science'
            Returns as science for most Exposures
        exposure_type : 'target_acquisition'
            Returns target_acquisition for mir_tacq
        """
        NEVER_CHANGE = ['target_acquisition']
        exp_type = super().get_exposure_type(item, default=default)
        if exp_type in NEVER_CHANGE:
            return exp_type
        return 'science'


class AsnMixin_Coronagraphy:
    """Basic overrides for Coronagraphy associations"""
    def __init__(self, *args, **kwargs):

        # PSF is required
        self.validity.update({
            'has_psf': {
                'validated': False,
                'check': lambda entry: entry['exptype'] == 'psf'
            }
        })

        super().__init__(*args, **kwargs)

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""
        self.data['asn_type'] = 'coron3'
        super()._init_hook(item)


class AsnMixin_Science(DMS_Level3_Base):
    """Basic science constraints"""

    def __init__(self, *args, **kwargs):

        # Setup target acquisition inclusion
        constraint_acqs = Constraint(
            [
                Constraint_TargetAcq(),
                DMSAttrConstraint(
                    name='acq_obsnum',
                    sources=['obs_num'],
                    value=lambda: '('
                    + '|'.join(self.constraints['obs_num'].found_values)
                    + ')',
                    force_unique=False,
                )
            ],
            name='acq_constraint',
            work_over=ListCategory.EXISTING
        )

        # Put all constraints together.
        self.constraints = Constraint(
            [
                Constraint_Base(),
                DMSAttrConstraint(
                    sources=['is_imprt'],
                    force_undefined=True
                ),
                Constraint(
                    [
                        Constraint(
                            [
                                self.constraints,
                                Constraint_Obsnum()
                            ],
                            name='rule'
                        ),
                        constraint_acqs
                    ],
                    name='acq_check',
                    reduce=Constraint.any
                ),
            ],
            name='dmsbase_top'
        )

        super(AsnMixin_Science, self).__init__(*args, **kwargs)

    def finalize(self):
        if self.acid.type.lower() in MULTI_OBS_AC_TYPES:
            if len(self.constraints['obs_num'].found_values) <= 1:
                return

        return super(AsnMixin_Science, self).finalize()


class AsnMixin_Spectrum(AsnMixin_Science):
    """All things that are spectrum"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        self.data['asn_type'] = 'spec3'
        super(AsnMixin_Spectrum, self)._init_hook(item)


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/pool.py
"""

from collections import UserDict

from astropy.io.ascii import convert_numpy
from astropy.table import Table

__all__ = ["AssociationPool"]

DEFAULT_DELIMITER = "|"
DEFAULT_FORMAT = "ascii"


class AssociationPool(Table):
    """Association Pool

    An ``AssociationPool`` is essentially an astropy Table with the
    following default behaviors:

    - ASCII tables with a default delimiter of `|`
    - All values are read in as strings
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta["pool_file"] = self.meta.get("pool_file", "in-memory")

    @classmethod
    def read(
        cls, filename, delimiter=DEFAULT_DELIMITER, format=DEFAULT_FORMAT, **kwargs
    ):
        """Read in a Pool file

        Parameters
        ----------
        filename : str
            File path to read in as a table.

        delimiter : str
            Character used to delineate columns.

        format : str
            The format of the input file.

        Returns
        -------
        AssociationPool
            The ``AssociationPool`` representation of the file.
        """
        table = super().read(
            filename,
            delimiter=delimiter,
            format=format,
            converters=convert_to_str,
            **kwargs,
        )

        # If anything has been masked, just fill
        table = table.filled("null")

        # Lowercase the column names
        # Note: Cannot do in-place because modifying the
        #       list while iterating.
        columns = [column for name, column in table.columns.items()]
        for c in columns:
            c.name = c.name.lower()

        table.meta["pool_file"] = filename
        return table

    def write(self, *args, **kwargs):
        """Write the pool to a file.

        Parameters
        ----------
        output : str, file-like
            The output file or file-like object.

        delimiter : str
            The string to use to delineate columns.
            Default is '|'.

        format : str
            The format the file should be written in.
            Default is 'ascii'.

        args, kwargs : obj
            Other parameters that ``astropy.io.ascii.write`` can accept.
        """
        delimiter = kwargs.pop("delimiter", DEFAULT_DELIMITER)
        format = kwargs.pop("format", DEFAULT_FORMAT)
        try:
            super().write(*args, delimiter=delimiter, format=format, **kwargs)
        except TypeError:
            # Most likely caused by the actual `write` called
            # does not handle `delimiter`. `jsviewer` is one
            # such format.
            # So, try again without a delimiter.
            super().write(*args, format=format, **kwargs)


class PoolRow(UserDict):
    """A row from an AssociationPool

    Class to create a copy of an AssociationPool row without copying
    all of the astropy.Table.Row private attributes.
    """

    def __init__(self, init=None):
        dict_init = dict(init)
        super().__init__(dict_init)
        try:
            self.meta = init.meta
        except AttributeError:
            self.meta = dict()


def _convert_to_str():
    func, type_ = convert_numpy(str)

    def convert_func(vals):
        """Lowercase the conversion"""
        results = func(vals)
        results = [result.lower() for result in results]
        return results

    return [(convert_func, type_)]


convert_to_str = {"*": _convert_to_str()}


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/tweakreg/tweakreg_step.py

JWST pipeline step for image alignment.

:Authors: Mihai Cara

"""
from os import path

from astropy.table import Table
from astropy.time import Time
from tweakwcs.correctors import JWSTWCSCorrector

import spike.stcal.tweakreg_tweakreg as twk

from spike.jwstcal.stpipe_utilities import record_step_status
from spike.jwstcal.assign_wcs_util import update_fits_wcsinfo, update_s_region_imaging
from spike.jwstcal.datamodels_library import ModelLibrary

# LOCAL
from spike.jwstcal.stpipe_core import JwstStep as Step
from spike.jwstcal.tweakreg_tweakreg_catalog import make_tweakreg_catalog


def _oxford_or_str_join(str_list):
    nelem = len(str_list)
    if not nelem:
        return 'N/A'
    str_list = list(map(repr, str_list))
    if nelem == 1:
        return str_list
    elif nelem == 2:
        return str_list[0] + ' or ' + str_list[1]
    else:
        return ', '.join(map(repr, str_list[:-1])) + ', or ' + repr(str_list[-1])


SINGLE_GROUP_REFCAT = ['GAIADR3', 'GAIADR2', 'GAIADR1']
_SINGLE_GROUP_REFCAT_STR = _oxford_or_str_join(SINGLE_GROUP_REFCAT)

__all__ = ['TweakRegStep']


class TweakRegStep(Step):
    """
    TweakRegStep: Image alignment based on catalogs of sources detected in
    input images.
    """

    class_alias = "tweakreg"

    spec = f"""
        save_catalogs = boolean(default=False) # Write out catalogs?
        use_custom_catalogs = boolean(default=False) # Use custom user-provided catalogs?
        catalog_format = string(default='ecsv') # Catalog output file format
        catfile = string(default='') # Name of the file with a list of custom user-provided catalogs
        starfinder = option('dao', 'iraf', 'segmentation', default='iraf') # Star finder to use.

        # general starfinder options
        snr_threshold = float(default=10.0) # SNR threshold above the bkg for star finder
        bkg_boxsize = integer(default=400) # The background mesh box size in pixels.

        # kwargs for DAOStarFinder and IRAFStarFinder, only used if starfinder is 'dao' or 'iraf'
        kernel_fwhm = float(default=2.5) # Gaussian kernel FWHM in pixels
        minsep_fwhm = float(default=0.0) # Minimum separation between detected objects in FWHM
        sigma_radius = float(default=1.5) # Truncation radius of the Gaussian kernel in units of sigma
        sharplo = float(default=0.2) # The lower bound on sharpness for object detection.
        sharphi = float(default=1.0) # The upper bound on sharpness for object detection.
        roundlo = float(default=-1.0) # The lower bound on roundness for object detection.
        roundhi = float(default=1.0) # The upper bound on roundness for object detection.
        brightest = integer(default=200) # Keep top ``brightest`` objects
        peakmax = float(default=None) # Filter out objects with pixel values >= ``peakmax``

        # kwargs for SourceCatalog and SourceFinder, only used if starfinder is 'segmentation'
        npixels = integer(default=10) # Minimum number of connected pixels
        connectivity = option(4, 8, default=8) # The connectivity defining the neighborhood of a pixel
        nlevels = integer(default=32) # Number of multi-thresholding levels for deblending
        contrast = float(default=0.001) # Fraction of total source flux an object must have to be deblended
        multithresh_mode = option('exponential', 'linear', 'sinh', default='exponential') # Multi-thresholding mode
        localbkg_width = integer(default=0) # Width of rectangular annulus used to compute local background around each source
        apermask_method = option('correct', 'mask', 'none', default='correct') # How to handle neighboring sources
        kron_params = float_list(min=2, max=3, default=None) # Parameters defining Kron aperture

        # align wcs options
        enforce_user_order = boolean(default=False) # Align images in user specified order?
        expand_refcat = boolean(default=False) # Expand reference catalog with new sources?
        minobj = integer(default=15) # Minimum number of objects acceptable for matching
        fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift') # Fitting geometry
        nclip = integer(min=0, default=3) # Number of clipping iterations in fit
        sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units

        # xyxymatch options
        searchrad = float(default=2.0) # The search radius in arcsec for a match
        use2dhist = boolean(default=True) # Use 2d histogram to find initial offset?
        separation = float(default=1.0) # Minimum object separation for xyxymatch in arcsec
        tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec
        xoffset = float(default=0.0), # Initial guess for X offset in arcsec
        yoffset = float(default=0.0) # Initial guess for Y offset in arcsec

        # Absolute catalog options
        abs_refcat = string(default='')  # Catalog file name or one of: {_SINGLE_GROUP_REFCAT_STR}, or None, or ''
        save_abs_catalog = boolean(default=False)  # Write out used absolute astrometric reference catalog as a separate product

        # Absolute catalog align wcs options
        abs_minobj = integer(default=15) # Minimum number of objects acceptable for matching when performing absolute astrometry
        abs_fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift')
        abs_nclip = integer(min=0, default=3) # Number of clipping iterations in fit when performing absolute astrometry
        abs_sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units when performing absolute astrometry

        # absolute catalog xyxymatch options
        abs_searchrad = float(default=6.0) # The search radius in arcsec for a match when performing absolute astrometry
        # We encourage setting this parameter to True. Otherwise, xoffset and yoffset will be set to zero.
        abs_use2dhist = boolean(default=True) # Use 2D histogram to find initial offset when performing absolute astrometry?
        abs_separation = float(default=1) # Minimum object separation in arcsec when performing absolute astrometry
        abs_tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec when performing absolute astrometry

        # SIP approximation options, should match assign_wcs
        sip_approx = boolean(default=True)  # enables SIP approximation for imaging modes.
        sip_max_pix_error = float(default=0.01)  # max err for SIP fit, forward.
        sip_degree = integer(max=6, default=None)  # degree for forward SIP fit, None to use best fit.
        sip_max_inv_pix_error = float(default=0.01)  # max err for SIP fit, inverse.
        sip_inv_degree = integer(max=6, default=None)  # degree for inverse SIP fit, None to use best fit.
        sip_npoints = integer(default=12)  #  number of points for SIP
        
        # stpipe general options
        output_use_model = boolean(default=True)  # When saving use `DataModel.meta.filename`
        in_memory = boolean(default=True)  # If False, preserve memory using temporary files at expense of runtime
    """

    reference_file_types: list = []

    def process(self, input):
        if isinstance(input, ModelLibrary):
            images = input
        else:
            images = ModelLibrary(input, on_disk=not self.in_memory)

        if len(images) == 0:
            raise ValueError("Input must contain at least one image model.")

        # determine number of groups (used below)
        n_groups = len(images.group_names)

        use_custom_catalogs = self.use_custom_catalogs

        if self.use_custom_catalogs:
            # first check catfile
            if self.catfile.strip():
                catdict = _parse_catfile(self.catfile)
                # if user requested the use of custom catalogs and provided a
                # valid 'catfile' file name that has no custom catalogs,
                # turn off the use of custom catalogs:
                if not catdict:
                    self.log.warning(
                        "'use_custom_catalogs' is set to True but 'catfile' "
                        "contains no user catalogs. Turning on built-in catalog "
                        "creation."
                    )
                    use_custom_catalogs = False
            # else, load from association
            elif images._asn_dir is not None:
                catdict = {}
                for member in images.asn["products"][0]["members"]:
                    if "tweakreg_catalog" in member:
                        tweakreg_catalog = member["tweakreg_catalog"]
                        if tweakreg_catalog is None or not tweakreg_catalog.strip():
                            catdict[member["expname"]] = None
                        else:
                            catdict[member["expname"]] = path.join(images._asn_dir, tweakreg_catalog)

        if self.abs_refcat is not None and self.abs_refcat.strip():
            align_to_abs_refcat = True
            # Set expand_refcat to True to eliminate possibility of duplicate
            # entries when aligning to absolute astrometric reference catalog
            self.expand_refcat = True
        else:
            align_to_abs_refcat = False

            # since we're not aligning to a reference catalog, check if we
            # are saving catalogs, if not, and we have 1 group, skip
            if not self.save_catalogs and n_groups == 1:
                # we need at least two exposures to perform image alignment
                self.log.warning("At least two exposures are required for image "
                                 "alignment.")
                self.log.warning("Nothing to do. Skipping 'TweakRegStep'...")
                record_step_status(images, "tweakreg", success=False)
                return images

        # === start processing images ===

        # pre-allocate collectors (same length and order as images)
        correctors = [None] * len(images)

        # Build the catalog and corrector for each input images
        with images:
            for (model_index, image_model) in enumerate(images):
                # now that the model is open, check its metadata for a custom catalog
                # only if it's not listed in the catdict
                if use_custom_catalogs and image_model.meta.filename not in catdict:
                    if (image_model.meta.tweakreg_catalog is not None and image_model.meta.tweakreg_catalog.strip()):
                        catdict[image_model.meta.filename] = image_model.meta.tweakreg_catalog
                if use_custom_catalogs and catdict.get(image_model.meta.filename, None) is not None:
                    image_model.meta.tweakreg_catalog = catdict[image_model.meta.filename]
                    # use user-supplied catalog:
                    self.log.info("Using user-provided input catalog "
                                  f"'{image_model.meta.tweakreg_catalog}'")
                    catalog = Table.read(
                        image_model.meta.tweakreg_catalog,
                    )
                    save_catalog = False
                else:
                    # source finding
                    catalog = self._find_sources(image_model)

                    # only save if catalog was computed from _find_sources and
                    # the user requested save_catalogs
                    save_catalog = self.save_catalogs

                # if needed rename xcentroid to x, ycentroid to y
                catalog = _rename_catalog_columns(catalog)

                # filter all sources outside the wcs bounding box
                catalog = twk.filter_catalog_by_bounding_box(
                    catalog,
                    image_model.meta.wcs.bounding_box)

                # setting 'name' is important for tweakwcs logging
                if catalog.meta.get('name') is None:
                    catalog.meta['name'] = path.splitext(image_model.meta.filename)[0].strip('_- ')

                # log results of source finding (or user catalog)
                filename = image_model.meta.filename
                nsources = len(catalog)
                if nsources == 0:
                    self.log.warning('No sources found in {}.'.format(filename))
                else:
                    self.log.info('Detected {} sources in {}.'
                                  .format(len(catalog), filename))

                # save catalog (if requested)
                if save_catalog:
                    # FIXME this modifies the input_model
                    image_model.meta.tweakreg_catalog = self._write_catalog(catalog, filename)

                # construct the corrector since the model is open (and already has a group_id)
                correctors[model_index] = \
                    twk.construct_wcs_corrector(image_model.meta.wcs,
                                                image_model.meta.wcsinfo.instance,
                                                catalog,
                                                image_model.meta.group_id,)
                images.shelve(image_model, model_index)

        self.log.info('')
        self.log.info("Number of image groups to be aligned: {:d}."
                      .format(n_groups))

        # wrapper to stcal tweakreg routines
        # step skip conditions should throw TweakregError from stcal
        if n_groups > 1:
            try:
            # relative alignment of images to each other (if more than one group)
                correctors = \
                    twk.relative_align(correctors,
                                       enforce_user_order=self.enforce_user_order,
                                       expand_refcat=self.expand_refcat,
                                       minobj=self.minobj,
                                       fitgeometry=self.fitgeometry,
                                       nclip=self.nclip,
                                       sigma=self.sigma,
                                       searchrad=self.searchrad,
                                       use2dhist=self.use2dhist,
                                       separation=self.separation,
                                       tolerance=self.tolerance,
                                       xoffset=self.xoffset,
                                       yoffset=self.yoffset,)
            except twk.TweakregError as e:
                self.log.warning(str(e))
                local_align_failed = True
            else:
                local_align_failed = False
        else:
            local_align_failed = True

        # absolute alignment to the reference catalog
        # can (and does) occur after alignment between groups
        if align_to_abs_refcat:
            with images:
                ref_image = images.borrow(0)
                try:
                    correctors = \
                        twk.absolute_align(correctors, self.abs_refcat,
                                        ref_wcs=ref_image.meta.wcs,
                                        ref_wcsinfo=ref_image.meta.wcsinfo.instance,
                                        epoch=Time(ref_image.meta.observation.date).decimalyear,
                                        abs_minobj=self.abs_minobj,
                                        abs_fitgeometry=self.abs_fitgeometry,
                                        abs_nclip=self.abs_nclip,
                                        abs_sigma=self.abs_sigma,
                                        abs_searchrad=self.abs_searchrad,
                                        abs_use2dhist=self.abs_use2dhist,
                                        abs_separation=self.abs_separation,
                                        abs_tolerance=self.abs_tolerance,
                                        save_abs_catalog=self.save_abs_catalog,
                                        abs_catalog_output_dir=self.output_dir,
                                                )
                    images.shelve(ref_image, 0, modify=False)
                except twk.TweakregError as e:
                    self.log.warning(str(e))
                    images.shelve(ref_image, 0, modify=False)
                    record_step_status(images, "tweakreg", success=False)
                    return images
                finally:
                    del ref_image

        if local_align_failed and not align_to_abs_refcat: 
            record_step_status(images, "tweakreg", success=False)
            return images

        # one final pass through all the models to update them based
        # on the results of this step
        self._apply_tweakreg_solution(images, correctors,
                                    align_to_abs_refcat=align_to_abs_refcat)
        return images


    def _apply_tweakreg_solution(self,
                        images: ModelLibrary,
                        correctors: list[JWSTWCSCorrector],
                        align_to_abs_refcat: bool = False,
                        ) -> ModelLibrary:
        with images:
            for (image_model, corrector) in zip(images, correctors):

                # retrieve fit status and update wcs if fit is successful:
                if ("fit_info" in corrector.meta and
                        "SUCCESS" in corrector.meta["fit_info"]["status"]):

                    # Update/create the WCS .name attribute with information
                    # on this astrometric fit as the only record that it was
                    # successful:
                    if align_to_abs_refcat:
                        # NOTE: This .name attrib agreed upon by the JWST Cal
                        #       Working Group.
                        #       Current value is merely a place-holder based
                        #       on HST conventions. This value should also be
                        #       translated to the FITS WCSNAME keyword
                        #       IF that is what gets recorded in the archive
                        #       for end-user searches.
                        corrector.wcs.name = f"FIT-LVL3-{self.abs_refcat}"

                    image_model.meta.wcs = corrector.wcs
                    update_s_region_imaging(image_model)

                    # Also update FITS representation in input exposures for
                    # subsequent reprocessing by the end-user.
                    if self.sip_approx:
                        try:
                            update_fits_wcsinfo(
                                image_model,
                                max_pix_error=self.sip_max_pix_error,
                                degree=self.sip_degree,
                                max_inv_pix_error=self.sip_max_inv_pix_error,
                                inv_degree=self.sip_inv_degree,
                                npoints=self.sip_npoints,
                                crpix=None
                            )
                        except (ValueError, RuntimeError) as e:
                            self.log.warning("Failed to update 'meta.wcsinfo' with FITS SIP "
                                            "approximation. Reported error is:")
                            self.log.warning(f'"{e.args[0]}"')
                record_step_status(image_model, "tweakreg", success=True)
                images.shelve(image_model)
        return images


    def _write_catalog(self, catalog, filename):
        '''
        Determine output filename for catalog based on outfile for step
        and output dir, then write catalog to file.

        Parameters
        ----------
        catalog : astropy.table.Table
            Table containing the source catalog.
        filename : str
            Output filename for step

        Returns
        -------
        catalog_filename : str
            Filename where the catalog was saved
        '''

        catalog_filename = str(filename).replace(
                    '.fits', '_cat.{}'.format(self.catalog_format)
                )
        if self.catalog_format == 'ecsv':
            fmt = 'ascii.ecsv'
        elif self.catalog_format == 'fits':
            # NOTE: The catalog must not contain any 'None' values.
            #       FITS will also not clobber existing files.
            fmt = 'fits'
        else:
            raise ValueError(
                '\'catalog_format\' must be "ecsv" or "fits".'
            )
        if self.output_dir is None:
            catalog.write(catalog_filename, format=fmt, overwrite=True)
        else:
            catalog.write(
                path.join(self.output_dir, catalog_filename),
                format=fmt,
                overwrite=True
            )
        self.log.info('Wrote source catalog: {}'
                      .format(catalog_filename))
        return catalog_filename

    def _find_sources(self, image_model):
        # source finding
        starfinder_kwargs = {
            'fwhm': self.kernel_fwhm,
            'sigma_radius': self.sigma_radius,
            'minsep_fwhm': self.minsep_fwhm,
            'sharplo': self.sharplo,
            'sharphi': self.sharphi,
            'roundlo': self.roundlo,
            'roundhi': self.roundhi,
            'peakmax': self.peakmax,
            'brightest': self.brightest,
            'npixels': self.npixels,
            'connectivity': int(self.connectivity),  # option returns a string, so cast to int
            'nlevels': self.nlevels,
            'contrast': self.contrast,
            'mode': self.multithresh_mode,
            'error': image_model.err,
            'localbkg_width': self.localbkg_width,
            'apermask_method': self.apermask_method,
            'kron_params': self.kron_params,
        }

        return make_tweakreg_catalog(
            image_model, self.snr_threshold,
            starfinder=self.starfinder,
            bkg_boxsize=self.bkg_boxsize,
            starfinder_kwargs=starfinder_kwargs,
        )


def _parse_catfile(catfile):
    """
    Parse a text file containing at 2 whitespace-delimited columns
        column 1: str, datamodel filename
        column 2: str, catalog filename
    into a dictionary with datamodel filename keys and catalog filename
    values. The catalog filenames will become paths relative
    to the current working directory. So for a catalog filename
    "mycat.ecsv" if the catfile is in a subdirectory "my_data"
    the catalog filename will be "my_data/mycat.ecsv".

    Returns:
        - None of catfile is None (or an empty string)
        - empty dict if catfile is empty

    Raises:
        VaueError if catfile contains >2 columns
    """
    if catfile is None or not catfile.strip():
        return None

    catdict = {}

    with open(catfile) as f:
        catfile_dir = path.dirname(catfile)

        for line in f.readlines():
            sline = line.strip()
            if not sline or sline[0] == '#':
                continue

            data_model, *catalog = sline.split()
            catalog = list(map(str.strip, catalog))
            if len(catalog) == 1:
                catdict[data_model] = path.join(catfile_dir, catalog[0])
            elif len(catalog) == 0:
                # set this to None so it's custom catalog is skipped
                catdict[data_model] = None
            else:
                raise ValueError("'catfile' can contain at most two columns.")

    return catdict


def _rename_catalog_columns(catalog):
    for axis in ['x', 'y']:
        if axis not in catalog.colnames:
            long_axis = axis + 'centroid'
            if long_axis in catalog.colnames:
                catalog.rename_column(long_axis, axis)
            else:
                raise ValueError(
                    "'tweakreg' source catalogs must contain either "
                    "columns 'x' and 'y' or 'xcentroid' and "
                    "'ycentroid'."
                )
    return catalog